from typing import Tuple, Dict
import time
from collections import defaultdict


KeyT = Tuple[int, str]


class Timer:
    state: str
    sum: int
    n: int

    def __init__(self, state: str):
        self.state = state
        self.sum = 0
        self.n = 0

    def __enter__(self):
        self.start = time.time()
        self.n += 1
        return self

    def __exit__(self, *args):
        interval = time.time() - self.start
        self.sum += interval


class ManyTimer:
    timers: Dict[str, Timer]

    def __init__(self):
        self.timers = {}

    def __call__(self, key: str) -> Timer:
        if key not in self.timers:
            self.timers[key] = Timer(key)
        return self.timers[key]


def set_params_proxy(model, proxy):
    """Set a 'proxy' on the internal ParamServer object for the model and
    its children. Experimental.
    """
    for node in model.walk():
        node._params.proxy = None
        for name in node.param_names:
            if node.has_param(name):
                proxy.set_param(node.id, name, node.get_param(name))
        node._params.proxy = proxy


def make_key(model_id: int, name: str) -> Tuple[int, str]:
    return (model_id, name)


def divide_params(model, num_workers):
    keys_by_node = defaultdict(list)
    for node in model.walk():
        keys = [make_key(node.id, name) for name in node.param_names]
        if keys:
            keys_by_node[node.id].extend(keys)
    key_groups = list(keys_by_node.values())
    n = max(1, len(key_groups) // num_workers)
    worker_keys = []
    start = 0
    for i in range(num_workers):
        worker_keys.append([])
        for kg in key_groups[start : start + n]:
            worker_keys[-1].extend(kg)
        start += n
    for kg in key_groups[start:]:
        worker_keys[-1].extend(kg)
    assert len(worker_keys) == num_workers, (len(worker_keys), num_workers)
    return worker_keys



from typing import Optional
import time
import typer
import logging
from pathlib import Path
from spacy.util import load_config, logger
from spacy.cli._util import parse_config_overrides, Arg, Opt, app
from spacy.cli._util import setup_gpu, show_validation_error
from thinc.api import Config

from .worker import Worker, Evaluator


RAY_HELP = """CLI for parallel and distributed computing via
Ray. See the Ray documentation for details: https://ray.io.
"""

# Create our subcommand, and install it within spaCy's CLI
ray_cli = typer.Typer(name="ray", help=RAY_HELP, no_args_is_help=True)
app.add_typer(ray_cli)


@ray_cli.command(
    "train", context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def ray_train_cli(
    # fmt: off
    ctx: typer.Context,  # This is only used to read additional arguments
    config_path: Path = Arg(..., help="Path to config file", exists=True),
    code_path: Optional[Path] = Opt(None, "--code", "-c", help="Path to Python file with additional code (registered functions) to be imported"),
    output_path: Optional[Path] = Opt(None, "--output", "--output-path", "-o", help="Output directory or remote storage URL for saving trained pipeline"),
    num_workers: int = Opt(1, "--n-workers", "-w", help="Number of workers"),
    ray_address: Optional[str] = Opt(None, "--address", "-a", help="Address of ray cluster"),
    use_gpu: int = Opt(-1, "--gpu-id", "-g", help="GPU ID or -1 for CPU"),
    verbose: bool = Opt(False, "--verbose", "-V", "-VV", help="Display more information for debugging purposes"),
    # fmt: on
):
    """
    Train a spaCy pipeline using Ray for parallel training.
    """
    # TODO: wire up output path
    logger.setLevel(logging.DEBUG if verbose else logging.ERROR)
    setup_gpu(use_gpu)
    overrides = parse_config_overrides(ctx.args)
    with show_validation_error(config_path):
        config = load_config(config_path, overrides=overrides, interpolate=False)
    ray_train(
        config,
        ray_address=ray_address,
        num_workers=num_workers,
        use_gpu=use_gpu,
        code_path=code_path,
    )


def ray_train(
    config: Config,
    *,
    ray_address: Optional[str] = None,
    num_workers: int = 1,
    use_gpu: int = -1,
    code_path: Optional[Path] = None,
) -> None:
    # We're importing Ray here so it doesn't need to be imported when spaCy /
    # spaCy's CLI is imported (which would otherwise take too long)
    import ray

    if ray_address is not None:
        ray.init(address=ray_address)
    else:
        ray.init(ignore_reinit_error=True)
    RemoteWorker = ray.remote(Worker).options(num_gpus=int(use_gpu >= 0), num_cpus=2)
    workers = [
        RemoteWorker.remote(
            config,
            rank=rank,
            num_workers=num_workers,
            use_gpu=use_gpu,
            code_path=code_path,
        )
        for rank in range(num_workers)
    ]
    for worker in workers:
        ray.get(worker.set_proxy.remote(workers))
    evaluator = ray.remote(Evaluator).remote()
    for worker in workers:
        ray.get(worker.train.remote(workers, evaluator))
    todo = list(workers)
    while todo:
        time.sleep(1)
        todo = [w for w in workers if ray.get(w.is_running.remote())]



import spacy
from spacy.language import Language

from . import mock_ray
from ..worker import Worker

# I don't normally go for so much mocking; I usually think it's pointless to
# test something so different. But this time it seems worth trying, just to
# get some basic tests, which would otherwise be hard.

# These tests currently really do nothing, I'm just laying out a skeleton while
# I think of what could be meaningfully tested this way. Maybe I'll delete it
# all in the end.


class TestWorker(Worker):
    def _load_nlp_and_config(self, config):
        return None, None

    def _initialize_models(self, nlp, config):
        return None, None


def test_worker_init():
    # Get a blank valid config
    nlp = spacy.blank("en")
    nlp.config["paths"]["train"] = ""
    nlp.config["paths"]["dev"] = ""
    worker = Worker(nlp.config, rank=1, num_workers=2, use_gpu=-1, ray=mock_ray)
    assert isinstance(worker.nlp, Language)



from typing import Dict, Any, Tuple, Callable
from datetime import timedelta
from spacy.util import registry
from spacy.errors import Errors
from wasabi import msg


@registry.loggers("spacy-ray.ConsoleLogger.v1")
def ray_console_logger():
    def setup_printer(
        nlp: "Language",
    ) -> Tuple[Callable[[Dict[str, Any]], None], Callable]:
        score_cols = list(nlp.config["training"]["score_weights"])
        score_widths = [max(len(col), 6) for col in score_cols]
        loss_cols = [f"Loss {pipe}" for pipe in nlp.pipe_names]
        loss_widths = [max(len(col), 8) for col in loss_cols]
        table_header = ["T", "E", "#", "W"] + loss_cols + score_cols + ["Score"]
        table_header = [col.upper() for col in table_header]
        table_widths = [8, 3, 6, 6] + loss_widths + score_widths + [6]
        table_aligns = ["r" for _ in table_widths]
        msg.row(table_header, widths=table_widths)
        msg.row(["-" * width for width in table_widths])

        def log_step(info: Dict[str, Any]):
            try:
                losses = [
                    "{0:.2f}".format(float(info["losses"][pipe_name]))
                    for pipe_name in nlp.pipe_names
                ]
            except KeyError as e:
                raise KeyError(
                    Errors.E983.format(
                        dict="scores (losses)",
                        key=str(e),
                        keys=list(info["losses"].keys()),
                    )
                ) from None

            try:
                scores = [
                    "{0:.2f}".format(float(info["other_scores"].get(col, 0.0)) * 100)
                    for col in score_cols
                ]
            except KeyError as e:
                raise KeyError(
                    Errors.E983.format(
                        dict="scores (other)",
                        key=str(e),
                        keys=list(info["other_scores"].keys()),
                    )
                ) from None
            time = timedelta(seconds=info["seconds"])
            data = (
                [str(time), info["epoch"], info["step"], info["words"]]
                + losses
                + scores
                + ["{0:.2f}".format(float(info["score"]))]
            )
            msg.row(data, widths=table_widths, aligns=table_aligns)

        def finalize():
            pass

        return log_step, finalize

    return setup_printer



from typing import Dict, Set, Iterable, Any, Optional, cast
from collections import Counter
from thinc.types import FloatsXd
from thinc.api import Optimizer

from .util import make_key, KeyT


class RayPeerProxy:
    """Proxy for workers where each worker owns some of the parameters. For
    parameters they don't own, workers will pull parameters and push gradients.
    For parameters they do own, they pull gradients, make the update, and push
    parameters.
    """

    ray: Any
    optimizer: Optimizer
    grads_per_update: int
    peers: Dict
    other_workers: Set
    _params: Dict[KeyT, FloatsXd]
    _grads: Dict[KeyT, Optional[FloatsXd]]
    _versions: Dict[KeyT, int]
    _owned_keys: Set[KeyT]
    _grad_counts: Dict[KeyT, int]

    def __init__(
        self,
        peers: Dict[KeyT, Any],
        optimizer,
        keys: Iterable[KeyT],
        *,
        grads_per_update: int = 2,
        ray=None
    ):
        if ray is None:
            import ray  # type: ignore
        # Pass in 'ray' so that we can test with a mock object.
        self.ray = ray
        self.optimizer = optimizer
        self.grads_per_update = grads_per_update
        self.peers = dict(peers)
        self._owned_keys = set(keys)
        self.other_workers = set()
        for key, peer in self.peers.items():
            if key not in self._owned_keys and peer not in self.other_workers:
                self.other_workers.add(peer)
        self._params = {}
        self._versions = Counter()
        self._next_params = {}
        self._grads = {}
        self._grad_counts = Counter()

    def check_version(self, key: KeyT, version: int) -> Optional[bool]:
        if key not in self._versions:
            return None
        elif self._versions[key] != version:
            return False
        else:
            return True

    def set_param(self, id, name, value: FloatsXd) -> None:
        """Set a parameter to the connection."""
        key = make_key(id, name)
        if key in self._owned_keys or key not in self._params:
            self._params[key] = value
            self._versions[key] += 1
            self._grads[key] = None
            self._grad_counts[key] = 0

    def send_param(self, key):
        param = self._params[key]
        version = self._versions[key]
        for peer in self.other_workers:
            peer.set_param.remote(key, version, param)

    def receive_param(self, key, version, value: FloatsXd) -> None:
        """Let the connection push a parameter to us."""
        # We have to store this in a separate place, to make sure we don't
        # fetch the wrong version when we submit the gradient. For instance,
        # imagine if we received the param in between the forward and backward
        # pass. If we set the version to this one, we'd calculate a gradient
        # on the basis of the old param, but think we had a new version.
        self._next_params[key] = (version, value)

    def get_param(self, id, name) -> FloatsXd:
        key = make_key(id, name)
        self._maybe_update_param(key)
        return self._params[key]

    def set_grad(self, id, name, value: FloatsXd) -> None:
        """Set a gradient to the connection."""
        key = make_key(id, name)
        if key in self._owned_keys:
            self._grads[key] = value
            self._grad_counts[key] = 1

    def inc_grad(self, id, name, value: FloatsXd) -> None:
        """Increment a gradient to the connection."""
        key = make_key(id, name)
        self._grad_counts[key] += 1
        if key not in self._owned_keys:
            peer = self.peers[key]
            peer.inc_grad.remote(key, self._versions[key], value)
        else:
            if self._grads.get(key) is None:
                self._grads[key] = value.copy()
            else:
                self._grads[key] += value

    def _maybe_update_param(self, key: KeyT) -> bool:
        if key in self._next_params:
            version, value = self._next_params.pop(key)
            self._params[key] = value
            self._versions[key] = version
            self._grad_counts[key] = 0
            self._grads[key] = None
            return True
        elif key not in self._owned_keys:
            return False
        elif self._grad_counts[key] < self.grads_per_update:
            return False
        elif self._grads.get(key) is None:
            return False
        else:
            grad = cast(FloatsXd, self._grads[key])
            self._versions[key] += 1
            param, _ = self.optimizer(key, self._params[key], grad)
            self._params[key] = param
            self._grads[key] = None
            self._grad_counts[key] = 0
            self.send_param(key)
            return True



from typing import List, Dict, Union, Any, Optional
import time
import os
import threading
from pathlib import Path
from thinc.config import Config
from thinc.types import FloatsXd
from spacy.cli._util import import_code
from spacy.training.loop import train_while_improving, create_train_batches
from spacy.training.loop import create_evaluation_callback
from spacy.training.loop import create_before_to_disk_callback
from spacy.training.loop import update_meta
from spacy.training.initialize import init_nlp
from spacy.language import Language
from spacy.util import registry, logger, resolve_dot_names
from spacy.schemas import ConfigSchemaTraining
from thinc.api import require_gpu, set_gpu_allocator

from .proxies import RayPeerProxy
from .util import set_params_proxy, divide_params, KeyT


class Worker:
    """Actor class for spaCy parallel training.

    Okay this is pretty mind-bending stuff...But the idea is that the remote
    workers need to communicate directly, to avoid extra copies. The mechanics
    of this are super twisted though, because it mixes all sorts of levels. But
    it has the be *exactly this* worker object that is passed through, because
    we need the remote actor. That's why it's so strange.

    The workers install a "proxy" object into the Thinc models. When this object
    is installed, Thinc will direct calls to get_param, inc_grad, set_param etc
    through the proxy.

    On each worker, a subset of the weights will be "local". The proxy will
    receive a list of the keys that are local, and a mapping of keys to workers.

    Workers optimize the parameters that are 'local' to them, and then push
    the updates to the other workers. For parameters that aren't local, they
    find the worker that owns that parameter, and publish the gradient to it.

    This strategy is non-blocking, because the gradients and the parameters
    are both pushed, not pulled.

    In order to make this work, we need some concurrency within the workers,
    because the workers need to be listening for updates while continuing to
    work. Currently this is implemented by putting the main training work
    on a thread, and letting the main thread continue to listen for connections.

    Finally, not that there's a pretty tangled circular reference here. I hate
    circular references, it makes the code hard to understand and makes
    Python use GC. But the circular reference here is necessary:

    * Workers hold a reference to the nlp object. Within the nlp object, models
        hold references to the "proxy" object.
    * The proxy object holds a reference to the peer mapping, whose values are
        the workers.
    """

    rank: int
    num_workers: int
    gpu_id: int
    nlp: Language
    config: Union[Dict[str, Any], Config]
    proxy: Optional[RayPeerProxy]
    thread: Optional[threading.Thread]
    _results: List
    _evaluation_callback: Any

    def __init__(
        self,
        config: Config,
        *,
        rank: int = 0,
        num_workers: int = 1,
        use_gpu: int = 0,
        code_path: Optional[Path] = None,
        ray=None,
    ):
        if ray is None:
            # Avoid importing ray in the module. This allows a test-ray to
            # be passed in, and speeds up the CLI.
            import ray  # type: ignore

            self.ray = ray
        import_code(code_path)
        self.rank = rank
        self.num_workers = num_workers
        self.gpu_id = self._resolve_gpu(use_gpu)
        self.nlp = init_nlp(Config(config), use_gpu=self.gpu_id)
        config = self.nlp.config.interpolate()
        self.T = registry.resolve(config["training"], schema=ConfigSchemaTraining)
        dot_names = [self.T["train_corpus"], self.T["dev_corpus"]]
        self.train_corpus, self.dev_corpus = resolve_dot_names(config, dot_names)
        self.before_to_disk = create_before_to_disk_callback(self.T["before_to_disk"])
        allocator = self.T["gpu_allocator"]
        if use_gpu >= 0 and allocator:
            set_gpu_allocator(allocator)
        self._evaluation_callback = lambda: {}
        self._results = []
        self._has_evaluation_callback = False
        self.thread = None
        self.proxy = None
        self.n_grads_used = 0
        self.n_grads_discarded = 0

    ########################################################################
    # Inter-worker communication
    #
    # It'd be nice to have this stuff in a different object, but we need
    # to pass the actual 'actor' handle around, we can't use a shared reference.
    # And if we made another actor, it would run within a different process.
    #
    #########################################################################

    def inc_grad(self, key: KeyT, version: int, value: FloatsXd) -> None:
        if self.proxy is None:
            raise ValueError("Proxy object not set")
        if self.proxy.check_version(key, version):
            self.proxy.inc_grad(key[0], key[1], value)

    def set_param(self, key: KeyT, version: int, value: FloatsXd) -> Optional[FloatsXd]:
        return self.proxy.receive_param(key, version, value)

    def get_param(self, key: KeyT, version: int) -> Optional[FloatsXd]:
        if self.proxy is None:
            raise ValueError("Proxy object not set")
        elif self.proxy.check_version(key, version):
            return self.proxy.get_param(key[0], key[1])
        else:
            return None

    #########################################################################
    # Process control. These are used by the script or function coordinating
    # the work.
    #
    ########################################################################

    def sync_params(self):
        for key in self.proxy._owned_keys:
            self.proxy.send_param(key)

    def get_percent_grads_used(self):
        total = self.n_grads_used + self.n_grads_discarded
        if total == 0:
            return None
        else:
            return self.n_grads_used / total

    def get_quorum(self) -> int:
        # Default to setting the 'quorum' to be the number of workers multiplied
        # by the accumulate_gradient value. This is how many gradients for a
        # parameter we will accumulate before running the optimizer.
        return self.num_workers * self.T["accumulate_gradient"]

    def train(self, peers: List, evaluator: "Evaluator") -> None:
        def evaluate():
            if self.rank == 0:
                scores = self.evaluate()
                self.ray.get(evaluator.set_scores.remote(scores))
                return scores
            else:
                scores = None
                while scores is None:
                    time.sleep(5)
                    scores = self.ray.get(evaluator.get_scores.remote())
                return scores

        train_batches = create_train_batches(
            self.nlp,
            self.train_corpus,
            self.T["batcher"],
            self.T["max_epochs"],
        )
        training_step_iterator = train_while_improving(
            self.nlp,
            FakeOptimizer(),
            train_batches,
            evaluate=evaluate,
            dropout=self.T["dropout"],
            accumulate_gradient=1,
            patience=self.T["patience"],
            max_steps=self.T["max_steps"],
            eval_frequency=self.T["eval_frequency"],
            exclude=self.T["frozen_components"],
            annotating_components=self.T["annotating_components"],
        )
        if self.rank == 0:
            print_row, finalize_logger = self.T["logger"](self.nlp)
        else:
            print_row = lambda: None
        self.thread = threading.Thread(
            target=thread_training,
            args=(
                training_step_iterator,
                print_row,
                self.rank,
                self.num_workers,
                self.gpu_id,
            ),
        )
        self.thread.start()

    def is_running(self):
        return self.thread.is_alive()

    def evaluate(self) -> Dict[str, Union[Dict[str, float], float]]:
        if not self._has_evaluation_callback:
            self._evaluation_callback = create_evaluation_callback(
                self.nlp,
                self.dev_corpus,
                self.T["score_weights"],
            )
            self._has_evaluation_callback = True
        return self._evaluation_callback()

    def save_checkpoint(self, info: Dict, output_path: Path) -> None:
        with self.nlp.select_pipes(disable=self.T["frozen_components"]):
            update_meta(self.T, self.nlp, info)
        self.before_to_disk(self.nlp).to_disk(output_path)

    def get_owned_keys(self):
        owned_keys = []
        for name, component in self.nlp.pipeline:
            if hasattr(component, "model"):
                worker_keys = divide_params(component.model, self.num_workers)
                owned_keys.extend(worker_keys[self.rank])
        return owned_keys

    def get_peer_map(self, workers):
        peer_map = {}
        for name, component in self.nlp.pipeline:
            if hasattr(component, "model"):
                worker_keys = divide_params(component.model, self.num_workers)
                for worker, keys in zip(workers, worker_keys):
                    for key in keys:
                        peer_map[key] = worker
        return peer_map

    def set_proxy(self, peers) -> None:
        proxy = RayPeerProxy(
            self.get_peer_map(peers),
            self.T["optimizer"],
            self.get_owned_keys(),
            ray=self.ray,
        )
        for name, component in self.nlp.pipeline:
            if hasattr(component, "model"):
                set_params_proxy(component.model, proxy)
        self.proxy = proxy

    def _resolve_gpu(self, use_gpu: int) -> int:
        if use_gpu >= 0:
            gpu_id = int(os.environ.get("CUDA_VISIBLE_DEVICES", -1))
            logger.info(f"Using GPU (isolated): {gpu_id}")
            require_gpu(0)
        else:
            logger.info("Using CPU")
            gpu_id = -1
        return gpu_id


class FakeOptimizer:
    def __init__(self):
        self.averages = {}

    def __call__(self, key, weights, gradient):
        # This shouldn't be called, because when we have the parameter proxy
        # installed, the gradients should never appear, and the `has_grad`
        # check in `model.finish_update` should return False.
        # However, it's difficult to guarantee that for all subclasses and shims
        # so it's safer to noop instead of raising.
        return weights, gradient

    def step_schedules(self):
        pass


class Evaluator:
    """Share evaluation results between workers.

    One worker should publish evaluation results to the evaluator,
    while the other workers should retrieve them (using a wait-loop if
    necessary).
    """

    def __init__(self):
        self.scores = []

    def set_scores(self, scores):
        self.scores.append(scores)
        return scores

    def get_scores(self):
        if not self.scores:
            return None
        else:
            return self.scores[-1]


def thread_training(training_step_iterator, print_row, rank, num_workers, gpu_id):
    if gpu_id >= 0:
        # I don't fully understand why we need to do this within the thread.
        # I think 0 is also correct here, because ray sets the available devices?
        require_gpu(0)
    for batch, info, is_best_checkpoint in training_step_iterator:
        if rank == 0 and is_best_checkpoint is not None:
            info["words"] *= num_workers
            print_row(info)
